use cfg_expr::{
    targets::{Arch, Family, Os, Triple, Vendor},
    Expression,
    Predicate,
    TargetPredicate,
};

pub const TARGET_ARCHES: [Arch; 7] = [
    Arch::x86_64,
    Arch::x86,
    Arch::aarch64,
    Arch::arm,
    Arch::powerpc64,
    Arch::s390x,
    Arch::riscv64,
];

pub const TARGET_TRIPLES: [Triple; 7] = [
    Triple::new_const("x86_64-unknown-linux-gnu"),
    Triple::new_const("i686-unknown-linux-gnu"),
    Triple::new_const("aarch64-unknown-linux-gnu"),
    Triple::new_const("armv7-unknown-linux-gnueabihf"),
    Triple::new_const("powerpc64le-unknown-linux-gnu"),
    Triple::new_const("s390x-unknown-linux-gnu"),
    Triple::new_const("riscv64gc-unknown-linux-gnu"),
];

pub fn applicable(expression: &Expression) -> bool {
    if expression.original() == "any()" {
        // empty "any()" predicates are used as "hacks" to improve cargo
        // dependency resolution and need to be kept
        return true;
    }

    TARGET_ARCHES
        .iter()
        .any(|target_arch| expression.eval(|pred| evaluate(pred, target_arch)))
}

pub fn evaluate(predicate: &Predicate, target_arch: &Arch) -> bool {
    use Predicate::*;

    match predicate {
        Target(target) => {
            use TargetPredicate as TP;
            match target {
                TP::Arch(arch) => {
                    let e = arch == target_arch;
                    log::debug!("eval: target_arch = {} (for {}) => {}", arch, target_arch, e);
                    e
                },
                TP::Endian(endian) => {
                    log::debug!("eval: target_endian = {:?} => true", endian);
                    true
                },
                TP::Env(env) => {
                    let e = ["", "gnu"].contains(&env.as_str());
                    log::debug!("eval: target_env = {} => {}", env, e);
                    e
                },
                TP::Family(family) => {
                    let e = family == &Family::unix;
                    log::debug!("eval: target_family = {} => {}", family, e);
                    e
                },
                TP::HasAtomic(atomic) => {
                    log::debug!("eval: target_has_atomic = {} => true", atomic);
                    true
                },
                TP::Os(os) => {
                    let e = os == &Os::linux;
                    log::debug!("eval: target_os = {} => {}", os, e);
                    e
                },
                TP::PointerWidth(pointer_width) => {
                    log::debug!("eval: target_pointer_width = {} => true", pointer_width);
                    true
                },
                TP::Vendor(vendor) => {
                    let e = vendor == &Vendor::unknown;
                    log::debug!("eval: target_vendor = {} => {}", vendor, e);
                    e
                },
                // ignore predicates that are not supported in Cargo.toml
                TP::Panic(panic) => {
                    log::debug!("eval: panic = {} => false", panic);
                    false
                },
                // ignore unstable attributes
                TP::Abi(_) => true,
            }
        },
        TargetFeature(_) => true,
        // ignore unknown bare predicates
        Flag(flag) => {
            log::debug!("eval: {} => false", flag);
            false
        },
        // ignore predicates that are not supported in Cargo.toml
        Test => {
            log::debug!("eval: test => false");
            false
        },
        DebugAssertions => {
            log::debug!("eval: debug_assertions => false");
            false
        },
        ProcMacro => {
            log::debug!("eval: proc_macro => false");
            false
        },
        Feature(feature) => {
            log::debug!("eval: feature = {} => false", feature);
            false
        },
        KeyValue { key, val } => {
            log::debug!("eval: {} = {} => false", key, val);
            false
        },
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn eval() {
        let test_cases = [
            ("cfg(any())", true),
            ("cfg(target_os = \"macos\")", false),
            ("cfg(any(foo, bar))", false),
            ("cfg(all(unix, target_pointer_width = \"16\"))", true),
            ("cfg(not(foo))", true),
            ("cfg(unix)", true),
            ("cfg(not(unix))", false),
            ("cfg(windows)", false),
            ("cfg(miri)", false),
            ("cfg(linux)", false),
            ("cfg(not(windows))", true),
            ("cfg(any(unix, windows))", true),
            ("cfg(any(windows, unix))", true),
            ("cfg(any(windows, windows, windows))", false),
            ("cfg(target_os = \"linux\")", true),
            ("cfg(any(target_os = \"linux\"))", true),
            ("cfg(all(target_os = \"linux\"))", true),
            ("cfg(any(target_os = \"linux\", target_os = \"macos\"))", true),
            ("cfg(any(target_pointer_width = \"16\", target_pointer_width = \"32\", target_pointer_width = \"64\"))", true),
            ("cfg(any(target_os = \"android\", target_os = \"linux\"))", true),
            // from rustix 0.36.8
            ("cfg(all(criterion, not(any(target_os = \"emscripten\", target_os = \"wasi\"))))", false),
            ("cfg(all(any(target_os = \"android\", target_os = \"linux\"), any(rustix_use_libc, miri, not(all(target_os = \"linux\", any(target_arch = \"x86\", all(target_arch = \"x86_64\", target_pointer_width = \"64\"), all(target_endian = \"little\", any(target_arch = \"arm\", all(target_arch = \"aarch64\", target_pointer_width = \"64\"), target_arch = \"powerpc64\", target_arch = \"riscv64\", target_arch = \"mips\", target_arch = \"mips64\"))))))))", true),
            ("cfg(all(not(rustix_use_libc), not(miri), target_os = \"linux\", any(target_arch = \"x86\", all(target_arch = \"x86_64\", target_pointer_width = \"64\"), all(target_endian = \"little\", any(target_arch = \"arm\", all(target_arch = \"aarch64\", target_pointer_width = \"64\"), target_arch = \"powerpc64\", target_arch = \"riscv64\", target_arch = \"mips\", target_arch = \"mips64\")))))", true),
            ("cfg(any(rustix_use_libc, miri, not(all(target_os = \"linux\", any(target_arch = \"x86\", all(target_arch = \"x86_64\", target_pointer_width = \"64\"), all(target_endian = \"little\", any(target_arch = \"arm\", all(target_arch = \"aarch64\", target_pointer_width = \"64\"), target_arch = \"powerpc64\", target_arch = \"riscv64\", target_arch = \"mips\", target_arch = \"mips64\")))))))", true),
            ("cfg(any(target_os = \"android\", target_os = \"linux\"))", true),
            // from tokio 1.21.2
            ("cfg(all(any(target_arch = \"wasm32\", target_arch = \"wasm64\"), not(target_os = \"wasi\")))", false),
            ("cfg(loom)", false),
            ("cfg(not(all(any(target_arch = \"wasm32\", target_arch = \"wasm64\"), target_os = \"unknown\")))", true),
            ("cfg(not(any(target_arch = \"wasm32\", target_arch = \"wasm64\")))", true),
            ("cfg(target_os = \"freebsd\")", false),
            ("cfg(tokio_unstable)", false),
        ];

        for (expr, expected) in test_cases {
            let cfg = Expression::parse(expr).unwrap();
            assert_eq!(applicable(&cfg), expected, "{}", expr);
        }
    }
}
