use std::borrow::Cow;
use std::{fs, io};

use clap::{Parser, Subcommand};

mod cfg;
mod normalize;
mod patch;
mod strip;

#[derive(Debug, Parser)]
#[command(disable_help_subcommand(true))]
struct Args {
    #[command(subcommand)]
    command: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// normalize crate version (i.e. drop +build metadata)
    NormalizeVersion {
        /// path to Cargo.toml file
        path: String,
        /// path to output file
        #[clap(short = 'o')]
        output: Option<String>,
        /// write patch directly
        #[clap(short = 'p')]
        patch: bool,
    },
    /// strip unused target-specific dependencies
    StripForeign {
        /// path to Cargo.toml file
        path: String,
        /// path to output file
        #[clap(short = 'o')]
        output: Option<String>,
        /// write patch directly
        #[clap(short = 'p')]
        patch: bool,
    },
}

fn main() -> anyhow::Result<()> {
    env_logger::builder()
        .format_timestamp(None)
        .format_module_path(false)
        .format_target(false)
        .filter_level(if cfg!(debug_assertions) {
            log::LevelFilter::Debug
        } else {
            log::LevelFilter::Info
        })
        .init();

    let args = Args::parse();

    match args.command {
        Command::NormalizeVersion { path, output, patch } => normalize_version(path, output, patch),
        Command::StripForeign { path, output, patch } => strip_foreign(path, output, patch),
    }
}

fn normalize_version(path: String, output: Option<String>, patch: bool) -> anyhow::Result<()> {
    let original = read_input(&path)?;
    let patched = normalize::normalize_version(&original)?;

    process_output(&original, patched.as_ref(), output.as_deref(), patch)
}

fn strip_foreign(path: String, output: Option<String>, patch: bool) -> anyhow::Result<()> {
    let original = read_input(&path)?;
    let patched = strip::strip_foreign(&original)?;

    process_output(&original, patched.as_ref(), output.as_deref(), patch)
}

fn process_output(original: &str, patched: &str, output: Option<&str>, patch: bool) -> anyhow::Result<()> {
    if original == patched {
        return Ok(());
    }

    let result = if patch {
        Cow::Owned(patch::make_patch(original, patched))
    } else {
        Cow::Borrowed(patched)
    };

    write_output(output, result.as_ref())?;

    Ok(())
}

fn read_input(path: &str) -> anyhow::Result<String> {
    if path == "-" {
        // read from stdin
        let mut buf = String::new();
        loop {
            let read = io::stdin().read_line(&mut buf)?;
            if read == 0 {
                break;
            }
        }
        Ok(buf)
    } else {
        // read file from path
        Ok(fs::read_to_string(path)?)
    }
}

fn write_output(output: Option<&str>, contents: &str) -> anyhow::Result<()> {
    if let Some(output) = output {
        fs::write(output, contents)?;
    } else {
        println!("{}", contents);
    }

    Ok(())
}
