use std::borrow::Cow;
use std::collections::HashSet;

use anyhow::anyhow;
use cfg_expr::{targets::Triple, Expression};
use toml_edit::Document;

use crate::cfg::{self, TARGET_TRIPLES};

fn get_known_target_triples() -> HashSet<Triple> {
    use cfg_expr::targets::ALL_BUILTINS;
    ALL_BUILTINS.iter().map(|t| t.triple.clone()).collect()
}

pub fn strip_foreign(original: &str) -> anyhow::Result<Cow<str>> {
    let mut toml: Document = original.parse()?;

    if toml.get("target").is_none() {
        // no "target" table: nothing to do
        return Ok(Cow::Borrowed(original));
    }

    // collect names of optional dependencies with and without associated implicit features
    let (_, implicit_optional) = get_optional_features_explicit_implicit(&mut toml)?;

    let target_item = toml
        .get_mut("target")
        .expect("Existence of target table was already checked");

    let Some(targets) = target_item.as_table_mut() else {
        return Err(anyhow!("Cargo.toml invalid: [target] is not a table"));
    };

    // remove non-applicable target-specific dependencies
    let removed_targets = remove_foreign_targets(targets);

    // collect names of removed dependencies and if they are optional
    let removed_with_optional = parse_removed_targets(removed_targets)?;

    // collect names of all remaining dependencies and if they are optional
    let remaining_with_optional = get_dependencies(&mut toml)?;

    // collect names of optional dependencies for which all occurrences were dropped
    let dropped_with_optional = get_dropped_dependencies(&removed_with_optional, &remaining_with_optional);

    // remove references to dropped dependencies from feature dependencies
    strip_dropped_features(&mut toml, dropped_with_optional, &implicit_optional)?;

    Ok(Cow::Owned(toml.to_string()))
}

fn remove_foreign_targets(targets: &mut toml_edit::Table) -> Vec<(String, toml_edit::Item)> {
    let known_target_triples = get_known_target_triples();

    // remove non-applicable target-specific dependencies
    let mut removed_targets = Vec::new();
    targets.retain(|key, item| {
        // check if key is a valid target triple
        // and check if key is a supported target
        let maybe_triple = Triple::new(key.to_owned());
        if known_target_triples.contains(&maybe_triple) {
            if TARGET_TRIPLES.contains(&maybe_triple) {
                return true;
            } else {
                log::info!("Removing dependencies for target: '{}'", key);
                removed_targets.push((key.to_owned(), item.clone()));
                return false;
            }
        }

        // check if key is a valid cfg-expression
        // and check if it is applicable on any supported target
        match Expression::parse(key) {
            Ok(expr) => {
                if cfg::applicable(&expr) {
                    true
                } else {
                    log::info!("Removing dependencies for target: '{}'", key);
                    removed_targets.push((key.to_owned(), item.clone()));
                    false
                }
            },
            Err(_) => {
                log::info!("Unknown target configuration: '{}'", key);
                removed_targets.push((key.to_owned(), item.clone()));
                false
            },
        }
    });

    removed_targets
}

fn parse_removed_targets(removed_targets: Vec<(String, toml_edit::Item)>) -> anyhow::Result<Vec<(String, bool)>> {
    let result = removed_targets
        .into_iter()
        .map(|(target_name, item)| {
            let Some(target) = item.as_table() else {
                return Err(anyhow!("Cannot parse target: {}", target_name));
            };

            let mut ret = Vec::new();

            // check dependencies
            if let Some(deps_item) = target.get("dependencies") {
                let Some(deps) = deps_item.as_table() else {
                    return Err(anyhow!("Cannot parse dependencies for target: {}", target_name));
                };

                for (dep_name, dep_item) in deps {
                    // plain string
                    if let Some(_version) = dep_item.as_str() {
                        ret.push((String::from(dep_name), false));
                        continue;
                    }

                    // table
                    if let Some(dep) = dep_item.as_table() {
                        if let Some(optional_item) = dep.get("optional") {
                            let Some(optional) = optional_item.as_bool() else {
                                return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                            };
                            ret.push((String::from(dep_name), optional));
                        } else {
                            ret.push((String::from(dep_name), false));
                        }
                        continue;
                    }

                    // inline table
                    if let Some(dep) = dep_item.as_inline_table() {
                        if let Some(optional_item) = dep.get("optional") {
                            let Some(optional) = optional_item.as_bool() else {
                                return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                            };
                            ret.push((String::from(dep_name), optional));
                        } else {
                            ret.push((String::from(dep_name), false));
                        }
                        continue;
                    }

                    return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                }
            }

            // check build-dependencies
            if let Some(deps_item) = target.get("build-dependencies") {
                let Some(deps) = deps_item.as_table() else {
                    return Err(anyhow!("Cannot parse build-dependencies for target: {}", target_name));
                };

                for (dep_name, dep_item) in deps {
                    // plain string
                    if let Some(_version) = dep_item.as_str() {
                        ret.push((String::from(dep_name), false));
                        continue;
                    }

                    // table
                    if let Some(dep) = dep_item.as_table() {
                        if let Some(optional_item) = dep.get("optional") {
                            let Some(optional) = optional_item.as_bool() else {
                                return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                            };
                            ret.push((String::from(dep_name), optional));
                        } else {
                            ret.push((String::from(dep_name), false));
                        }
                        continue;
                    }

                    // inline table
                    if let Some(dep) = dep_item.as_inline_table() {
                        if let Some(optional_item) = dep.get("optional") {
                            let Some(optional) = optional_item.as_bool() else {
                                return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                            };
                            ret.push((String::from(dep_name), optional));
                        } else {
                            ret.push((String::from(dep_name), false));
                        }
                        continue;
                    }

                    return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                }
            }

            // dev-dependencies cannot be optional or referenced by features
            Ok(ret)
        })
        .collect::<anyhow::Result<Vec<Vec<(String, bool)>>>>()?
        .into_iter()
        .flatten()
        .collect();

    Ok(result)
}

fn get_dependencies(toml: &mut Document) -> anyhow::Result<HashSet<(String, bool)>> {
    let target_item = toml
        .get_mut("target")
        .expect("Existence of target table was already checked");
    let Some(targets) = target_item.as_table_mut() else {
        return Err(anyhow!("Cargo.toml invalid: [target] is not a table"));
    };

    let mut remaining_with_optional = HashSet::new();

    for (target_name, item) in targets.iter() {
        let Some(target) = item.as_table() else {
            return Err(anyhow!("Cannot parse target: {}", target_name));
        };

        if let Some(deps_item) = target.get("dependencies") {
            let Some(deps) = deps_item.as_table() else {
                return Err(anyhow!("Cannot parse dependencies for target: {}", target_name));
            };

            for (dep_name, dep_item) in deps {
                // plain string
                if let Some(_version) = dep_item.as_str() {
                    remaining_with_optional.insert((String::from(dep_name), false));
                    continue;
                }

                // table
                if let Some(dep) = dep_item.as_table() {
                    if let Some(optional_item) = dep.get("optional") {
                        let Some(optional) = optional_item.as_bool() else {
                            return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                        };
                        remaining_with_optional.insert((String::from(dep_name), optional));
                    } else {
                        remaining_with_optional.insert((String::from(dep_name), false));
                    }
                    continue;
                }

                // inline table
                if let Some(dep) = dep_item.as_inline_table() {
                    if let Some(optional_item) = dep.get("optional") {
                        let Some(optional) = optional_item.as_bool() else {
                            return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                        };
                        remaining_with_optional.insert((String::from(dep_name), optional));
                    } else {
                        remaining_with_optional.insert((String::from(dep_name), false));
                    }
                    continue;
                }

                return Err(anyhow!("Cannot parse dependency: {}", dep_name));
            }
        }

        if let Some(deps_item) = target.get("build-dependencies") {
            let Some(deps) = deps_item.as_table() else {
                return Err(anyhow!("Cannot parse build-dependencies for target: {}", target_name));
            };

            for (dep_name, dep_item) in deps {
                // plain string
                if let Some(_version) = dep_item.as_str() {
                    remaining_with_optional.insert((String::from(dep_name), false));
                    continue;
                }

                // table
                if let Some(dep) = dep_item.as_table() {
                    if let Some(optional_item) = dep.get("optional") {
                        let Some(optional) = optional_item.as_bool() else {
                            return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                        };
                        remaining_with_optional.insert((String::from(dep_name), optional));
                    } else {
                        remaining_with_optional.insert((String::from(dep_name), false));
                    }
                    continue;
                }

                // inline table
                if let Some(dep) = dep_item.as_inline_table() {
                    if let Some(optional_item) = dep.get("optional") {
                        let Some(optional) = optional_item.as_bool() else {
                            return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                        };
                        remaining_with_optional.insert((String::from(dep_name), optional));
                    } else {
                        remaining_with_optional.insert((String::from(dep_name), false));
                    }
                    continue;
                }

                return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
            }
        }
    }

    if let Some(deps_item) = toml.get("dependencies") {
        let Some(deps) = deps_item.as_table() else {
            return Err(anyhow!("Cannot parse dependencies."));
        };

        for (dep_name, dep_item) in deps {
            // plain string
            if let Some(_version) = dep_item.as_str() {
                remaining_with_optional.insert((String::from(dep_name), false));
                continue;
            }

            // table
            if let Some(dep) = dep_item.as_table() {
                if let Some(optional_item) = dep.get("optional") {
                    let Some(optional) = optional_item.as_bool() else {
                        return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                    };
                    remaining_with_optional.insert((String::from(dep_name), optional));
                } else {
                    remaining_with_optional.insert((String::from(dep_name), false));
                }
                continue;
            }

            // inline table
            if let Some(dep) = dep_item.as_inline_table() {
                if let Some(optional_item) = dep.get("optional") {
                    let Some(optional) = optional_item.as_bool() else {
                        return Err(anyhow!("Cannot parse dependency: {}", dep_name));
                    };
                    remaining_with_optional.insert((String::from(dep_name), optional));
                } else {
                    remaining_with_optional.insert((String::from(dep_name), false));
                }
                continue;
            }

            return Err(anyhow!("Cannot parse dependency: {}", dep_name));
        }
    }

    if let Some(deps_item) = toml.get("build-dependencies") {
        let Some(deps) = deps_item.as_table() else {
            return Err(anyhow!("Cannot parse build-dependencies."));
        };

        for (dep_name, dep_item) in deps {
            // plain string
            if let Some(_version) = dep_item.as_str() {
                remaining_with_optional.insert((String::from(dep_name), false));
                continue;
            }

            // table
            if let Some(dep) = dep_item.as_table() {
                if let Some(optional_item) = dep.get("optional") {
                    let Some(optional) = optional_item.as_bool() else {
                        return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                    };
                    remaining_with_optional.insert((String::from(dep_name), optional));
                } else {
                    remaining_with_optional.insert((String::from(dep_name), false));
                }
                continue;
            }

            // inline table
            if let Some(dep) = dep_item.as_inline_table() {
                if let Some(optional_item) = dep.get("optional") {
                    let Some(optional) = optional_item.as_bool() else {
                        return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
                    };
                    remaining_with_optional.insert((String::from(dep_name), optional));
                } else {
                    remaining_with_optional.insert((String::from(dep_name), false));
                }
                continue;
            }

            return Err(anyhow!("Cannot parse build-dependency: {}", dep_name));
        }
    }

    Ok(remaining_with_optional)
}

fn get_optional_features_explicit_implicit(toml: &mut Document) -> anyhow::Result<(HashSet<String>, HashSet<String>)> {
    let dependencies = get_dependencies(toml)?;
    let optional_dependencies: HashSet<String> = dependencies
        .into_iter()
        .filter(|(_, optional)| *optional)
        .map(|(dep_name, _)| dep_name)
        .collect();

    let mut feature_names: HashSet<String> = HashSet::new();
    if let Some(features_item) = toml.get_mut("features") {
        let Some(features) = features_item.as_table_mut() else {
            return Err(anyhow!("Cannot parse features."));
        };

        for (feature_name, _) in features.iter_mut() {
            feature_names.insert(feature_name.to_string());
        }
    }

    let mut explicit_optional: HashSet<String> = HashSet::new();
    if let Some(features_item) = toml.get_mut("features") {
        let Some(features) = features_item.as_table_mut() else {
            return Err(anyhow!("Cannot parse features."));
        };

        for (feature_name, feature_item) in features.iter_mut() {
            let Some(feature_deps_array) = feature_item.as_array_mut() else {
                return Err(anyhow!(
                    "Cannot parse feature dependencies for feature: {}",
                    feature_name
                ));
            };

            for feature_dep in feature_deps_array.iter() {
                let Some(feature_dep_name) = feature_dep.as_str() else {
                    return Err(anyhow!(
                        "Cannot parse feature dependencies for feature: {}",
                        feature_name
                    ));
                };

                if let Some(dep) = feature_dep_name.strip_prefix("dep:") {
                    explicit_optional.insert(String::from(dep));
                }
            }
        }
    }

    let implicit_optional: HashSet<String> = optional_dependencies
        .iter()
        .filter(|dep| !explicit_optional.contains(*dep))
        .cloned()
        .collect();

    Ok((explicit_optional, implicit_optional))
}

fn get_dropped_dependencies<'a>(
    removed_with_optional: &'a [(String, bool)],
    remaining_with_optional: &HashSet<(String, bool)>,
) -> Vec<&'a (String, bool)> {
    removed_with_optional
        .iter()
        .filter(|(dep_name, _optional)| {
            !remaining_with_optional.contains(&(dep_name.to_owned(), true))
                && !remaining_with_optional.contains(&(dep_name.to_owned(), false))
        })
        .collect()
}

fn strip_dropped_features(
    toml: &mut Document,
    dropped_with_optional: Vec<&(String, bool)>,
    implicit_optional: &HashSet<String>,
) -> anyhow::Result<()> {
    let mut dropped_features: HashSet<String> = HashSet::new();

    if let Some(features_item) = toml.get_mut("features") {
        let Some(features) = features_item.as_table_mut() else {
            return Err(anyhow!("Cannot parse features."));
        };

        for (feature_name, feature_item) in features.iter_mut() {
            let Some(feature_deps_array) = feature_item.as_array_mut() else {
                return Err(anyhow!(
                    "Cannot parse feature dependencies for feature: {}",
                    feature_name
                ));
            };

            let feature_deps_len = feature_deps_array.len();
            let mut removed_feat_dep = Vec::new();

            for feature_dep in feature_deps_array.iter() {
                let Some(feature_dep_name) = feature_dep.as_str() else {
                    return Err(anyhow!(
                        "Cannot parse feature dependencies for feature: {}",
                        feature_name
                    ));
                };

                if implicit_optional.contains(feature_dep_name)
                    && dropped_with_optional.contains(&&(String::from(feature_dep_name), true))
                {
                    removed_feat_dep.push(String::from(feature_dep_name));
                }

                if let Some(dep) = feature_dep_name.strip_prefix("dep:") {
                    if dropped_with_optional.contains(&&(String::from(dep), true)) {
                        removed_feat_dep.push(String::from(feature_dep_name));

                        if feature_name == dep && feature_deps_len == 1 {
                            dropped_features.insert(feature_name.to_string());
                        }
                    }
                    continue;
                }

                if let Some((dep, _feat)) = feature_dep_name.split_once("?/") {
                    if dropped_with_optional.contains(&&(String::from(dep), true)) {
                        removed_feat_dep.push(String::from(feature_dep_name));
                    }
                    continue;
                }

                if let Some((dep, _feat)) = feature_dep_name.split_once('/') {
                    if dropped_with_optional.contains(&&(String::from(dep), true))
                        || dropped_with_optional.contains(&&(String::from(dep), false))
                    {
                        removed_feat_dep.push(String::from(feature_dep_name));
                    }
                    continue;
                }
            }

            feature_deps_array.retain(|n| !removed_feat_dep.contains(&String::from(n.as_str().unwrap())));
        }

        features.retain(|key, _value| !dropped_features.contains(key));
    }

    Ok(())
}
